import atexit
import csv
import json
import logging
import logging.handlers as log_handlers
import shutil
import sys
import threading
import time
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Literal
from urllib.parse import urlencode

import requests

try:
    # todo: upgrade https://gpiozero.readthedocs.io/en/stable/migrating_from_rpigpio.html
    # checks if you have access to RPi.GPIO, which is available inside RPi
    import RPi.GPIO as GPIO

except (ImportError, RuntimeError):
    # for running in dev machines
    from Mock import GPIO

import os
import warnings

from apscheduler.schedulers.background import BackgroundScheduler
from flask import Flask, request
from pydantic import BaseModel, Field, SecretStr

# config variables
CONFIG_DIR = Path(__file__).parent / "configDir"
LOG_PATH = CONFIG_DIR / "parkpow.log"
app = Flask(__name__)


class ConfigData(BaseModel):
    """Configuration for the gate opener"""

    class Gate(BaseModel):
        """Gate configuration"""

        cameras: list[str]
        tags: list[str]
        io_pin: int = Field(alias="io-pin")
        orientation: Literal["front", "rear"] | None = None

    pp_update_from: bool = Field(alias="pp-update-from", default=True)
    gate_open_period: int = Field(alias="gate-open-period", default=30)
    poll_frequency: int = Field(alias="poll-frequency", default=20)
    pp_api_token: SecretStr = Field(alias="pp-api-token")
    activity_io_pin: int | None = Field(alias="activity-io-pin", default=None)
    gates: list[Gate] = Field(default_factory=list)

    def out_activity(self, val: bool) -> None:
        """just a wrapper that can handle additional io-pin used for activity and not related to gates"""
        if self.activity_io_pin:
            GPIO.output(self.activity_io_pin, val)

    def start_progress(self):
        self.out_activity(True)

    def stop_progress(self):
        self.out_activity(False)

    def get_tag_filter(self):
        tags = []
        has_empty_tags = False
        for gate in self.gates:
            if gate.tags:
                tags.extend(gate.tags)
            else:
                has_empty_tags = True

        if tags:
            # no need to apply tags filter. some gates have no tags specified
            if has_empty_tags:
                return
            yield "tags", ",".join(tags)
        else:
            yield "tags__isnull", True

    @classmethod
    def load_config(cls, path="config.json") -> "ConfigData":
        configLocation = CONFIG_DIR / path
        # ------------------------Get Configuration Data---------------------#
        if not configLocation.exists():
            raise Exception(
                f"`{configLocation}` is not found. "
                "Please create one from `configDir/config.sample.json`"
            )

        try:
            return cls.parse_file(configLocation)
        except Exception as ex:
            sys.exit(
                f"Failed to load config file({configLocation}). Please check its format: {ex}"
            )


# logging settings
def log_file_namer(default_name):
    # This will be called when doing the log rotation
    # default_name is the default filename that would be assigned, e.g. Rotate_Test.txt.YYYY-MM-DD
    # Do any manipulations to that name here, for example this changes the name to Rotate_Test.YYYY-MM-DD.txt
    base_filename, ext, date = default_name.split(".")
    return f"{base_filename}.{date}.{ext}"


def get_logger():
    formatter = logging.Formatter("%(asctime)s:%(levelname)s:%(message)s")
    console = logging.StreamHandler()
    console.setFormatter(formatter)
    file_log = log_handlers.TimedRotatingFileHandler(
        filename=str(LOG_PATH),
        when="midnight",
        interval=1,
        backupCount=20,
    )
    file_log.setFormatter(formatter)
    file_log.namer = log_file_namer
    logger = logging.getLogger()
    logger.addHandler(console)
    logger.addHandler(file_log)
    logger.setLevel(int(os.getenv("LOG_LEVEL", 20)))
    return logger


logger = get_logger()

warnings.filterwarnings("ignore", message=".*pytz.*")


class AccessFile:
    """encapsulate accessList.csv functions"""

    def __init__(self):
        self.mtime = 0.0
        self.cache = None
        self.path = CONFIG_DIR / "accessList.csv"

    def update(self):
        # --------Regularly called to update plate access list---------------#
        CONFIG.start_progress()
        url = "https://app.parkpow.com/api/v1/vehicles/"

        logger.info("Access List Update Start")

        # ------------Import JSON data into CSV-----------#
        # create new file so that, the web-hook can access old version in the meantime
        new_file = self.path.with_name("access-list.new.csv")
        headers = {"Authorization": f"Token {CONFIG.pp_api_token.get_secret_value()}"}
        with new_file.open("w") as fw:
            access_list_writer = csv.DictWriter(
                fw, fieldnames=FIELDS, extrasaction="ignore"
            )
            kwargs = {"page": 1}
            kwargs.update(CONFIG.get_tag_filter())
            next_page = url + "?" + urlencode(kwargs)
            while next_page is not None:
                try:
                    json_data_curl = requests.get(
                        next_page, headers=headers, timeout=60
                    ).json()
                except Exception as ex:
                    logger.error(f"Update Access List Failed: {ex}")
                    break

                if "results" not in json_data_curl:
                    logger.error(
                        f"Update Access List failed: Wrong data-format received. {json_data_curl=}"
                    )
                    break

                next_page = json_data_curl.get("next")
                plate_list = json_data_curl["results"]
                for plate in plate_list:
                    access_list_writer.writerow(plate)
        shutil.move(new_file, self.path)
        time.sleep(1)
        CONFIG.stop_progress()

    def get_row(self, plate: str):
        mtime = os.path.getmtime(self.path)
        if self.mtime != mtime:
            # Find a match in the CSV
            with open(self.path) as fp:
                self.mtime = mtime
                self.cache = {
                    row["license_plate"].lower().strip(): row
                    for row in csv.DictReader(fp, fieldnames=FIELDS)
                }

        return self.cache.get(plate.lower().strip())


CONFIG = ConfigData.load_config()
"""config singleton"""
ACCESS_FILE = AccessFile()

FIELDS = [
    "license_plate",
    "payment_status",
    "make",
    "model",
    "color",
    "type",
    "field1",
    "field2",
    "field3",
    "field4",
    "field5",
    "field6",
    "vehicle_tags",
]


def configure_io_pins():
    GPIO.setwarnings(False)
    GPIO.setmode(GPIO.BCM)

    for gate in CONFIG.gates:
        GPIO.setup(gate.io_pin, GPIO.OUT)

    if CONFIG.activity_io_pin:
        GPIO.setup(CONFIG.activity_io_pin, GPIO.OUT)

    for gate in CONFIG.gates:
        GPIO.output(gate.io_pin, True)

    CONFIG.start_progress()
    time.sleep(1)
    CONFIG.stop_progress()


def schedule_task():
    # ---Calls "updateAccessList", pollFrequency comes from config.ini---#
    if (str(CONFIG.pp_update_from).lower().strip() not in {"true", "1", "yes"}) or (
        CONFIG.poll_frequency == 0
    ):
        logger.info("Auto updating of ACL not enabled")
        return
    logger.info("Auto updating via ParkPow API Enabled")
    scheduler = BackgroundScheduler()
    scheduler.add_job(
        func=ACCESS_FILE.update, trigger="interval", seconds=CONFIG.poll_frequency
    )
    scheduler.start()
    atexit.register(lambda: scheduler.running and scheduler.shutdown())
    return scheduler


def open_close_gate(pin_num):
    GPIO.output(pin_num, False)
    time.sleep(CONFIG.gate_open_period)
    GPIO.output(pin_num, True)


def get_vehicle_tags(plate: str):
    # Find a match in the CSV
    access_item = ACCESS_FILE.get_row(plate)
    if access_item is None:
        logger.info(f"License plate {plate} is not in the access list.")
        return None

    vehicle_tags = access_item["vehicle_tags"]
    try:
        return json.loads(vehicle_tags.replace("'", '"')) if vehicle_tags else []
        # empty list if there are no tags
    except JSONDecodeError:
        logger.error(f"Tags {access_item['vehicle_tags']} are not correctly formatted.")
    return None


def get_vehicle_orientation(result: dict) -> "str|None":
    orientation: "str | None" = None
    if "orientation" in result:
        orients = result.get("orientation")
    else:
        orients = result.get("vehicle", {}).get("props", {}).get("orientation", ())

    if orients:
        ont = orients[0]
        orientation = ont["value"] if "value" in ont else ont.get("orientation")
    if orientation:
        orientation = orientation.lower()
    if orientation == "unknown":
        orientation = None
    return orientation


def handle_visit_data(camera_id, result: dict) -> tuple[str, str]:
    # the format is same as the one in visit-API
    stream_plate = result["plate"]
    tags = get_vehicle_tags(stream_plate)
    if tags is None:
        return stream_plate, "Error getting vehicle-tags"

    # Find the matching gate
    orientation = get_vehicle_orientation(result)
    for idx, gate in enumerate(CONFIG.gates):
        tags_ok = (not gate.tags) or (set(gate.tags) & set(tags))
        orientation_ok = (not gate.orientation) or (gate.orientation == orientation)
        if tags_ok and (camera_id in gate.cameras) and orientation_ok:
            logger.info(
                f"Plate {stream_plate} accepted by {camera_id} through gate {idx + 1}"
            )
            pin_num = gate.io_pin

            # control io-pins in background and do not halt the web-server
            threading.Thread(target=open_close_gate, args=[pin_num]).start()
            break
    else:
        logger.warning(
            f"Camera {camera_id} with plate {stream_plate} does not match any gate."
        )

    return stream_plate, "JSON Posted"


@app.route("/postJson", methods=["POST"])
def json_handler():
    stream_data = json.loads(request.form["json"])
    response = []

    camera_id = stream_data["hook"]["id"]
    for result in stream_data["data"]["results"]:
        plate, message = handle_visit_data(camera_id, result)
        response.append({"plate": plate, "message": message})

    return response


def main():
    logger.info("Parkpow Server Started")
    logger.debug(f"{CONFIG.gates=}")
    configure_io_pins()
    schedule_task()
    app.run(host="0.0.0.0", port=8010)  # noqa


if __name__ == "__main__":
    main()
