import csv
import json
import time
from unittest.mock import MagicMock

import pytest
from parkpow_gate_opener import ACCESS_FILE, FIELDS


@pytest.fixture(autouse=True)
def clean_cache():
    ACCESS_FILE.mtime = None


@pytest.fixture
def mock_access_list(fixture_path, monkeypatch):
    file = fixture_path / "accessList.csv"
    monkeypatch.setattr(ACCESS_FILE, "path", file)


def test_json_handler(monkeypatch, mock_access_list, client):
    from parkpow_gate_opener import CONFIG, GPIO

    monkeypatch.setattr(CONFIG, "gate_open_period", 1)
    gpio_out = MagicMock()
    monkeypatch.setattr(GPIO, "output", gpio_out)

    stream_data = json.dumps(
        dict(hook=dict(id="gate1_cam2"), data=dict(results=[dict(plate="54-1263R")]))
    )
    response = client.post("/postJson", data=dict(json=stream_data))
    assert response.status_code == 200

    # wait for threads to finish
    time.sleep(2)

    # check that the io-pins are set
    assert gpio_out.call_count == 2


@pytest.fixture
def mock_api_response(mocked_responses, fixture_path):
    file = fixture_path / "vehicles-api.json"
    mocked_responses.get(
        "https://app.parkpow.com/api/v1/vehicles/",
        json=json.loads(file.read_text()),
        status=200,
    )


def test_access_list_update(monkeypatch, tmp_path, mock_api_response):
    file = tmp_path / "file.csv"
    monkeypatch.setattr(ACCESS_FILE, "path", file)

    ACCESS_FILE.update()

    assert ACCESS_FILE.path.exists()

    with open(ACCESS_FILE.path) as fp:
        rows = list(csv.DictReader(fp, fieldnames=FIELDS))

    assert len(rows) == 3
    row = ACCESS_FILE.get_row("KEU 147")
    assert row["vehicle_tags"] == "['Authorized']"


def test_scheduled_call(gpio):
    from parkpow_gate_opener import schedule_task

    scheduler = schedule_task()
    # some minimal test that checks that modules is loading
    assert len(scheduler.get_jobs()) == 1
    scheduler.shutdown()


def test_gpio_installed():
    import RPi.GPIO as gpio

    # will raise if the container is not started properly
    # RuntimeError: This module can only be run on a Raspberry Pi!
    assert gpio.VERSION
